"""Test the TianPwr BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.tianpwr_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests


def ref_value() -> BMSSample:
    """Return reference value for mock Tian Power BMS."""
    return {
        "temp_sensors": 4,
        "voltage": 54.74,
        "current": 0.0,
        "battery_level": 60,
        "battery_health": 100,
        "cycle_charge": 138.96,
        "cycles": 0,
        "temperature": 24.167,
        "cycle_capacity": 7606.67,
        "power": 0.0,
        "design_capacity": 230,
        "battery_charging": False,
        "cell_count": 16,
        "cell_voltages": [
            3.415,
            3.419,
            3.419,
            3.422,
            3.421,
            3.422,
            3.425,
            3.424,
            3.421,
            3.425,
            3.421,
            3.421,
            3.422,
            3.42,
            3.422,
            3.429,
        ],
        "temp_values": [28.0, 25.0, 23.0, 23.0, 23.0, 23.0],
        "delta_voltage": 0.014,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "problem": False,
        "problem_code": 0,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockTianPwrBleakClient(MockBleakClient):
    """Emulate a TianPwr BMS BleakClient."""

    RESP: Final[dict[int, bytearray]] = {
        0x81: bytearray(  # Software version frame
            b"\x55\x14\x81\x30\x2e\x31\x2e\x31\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x82: bytearray(  # Hardware version frame
            b"\x55\x14\x82\x54\x50\x2d\x4c\x54\x35\x35\x00\x54\x42\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x83: bytearray(  # Status frame
            b"\x55\x14\x83\x00\x3c\x15\x62\x01\x18\x00\xe6\x00\xfa\x00\x00\x30\x30\x00\x64\xaa"
        ),  # 60%, 54.74V, 0A, 28° ambient temp, 23°, 25° MOS temp
        0x84: bytearray(  # General info frame
            b"\x55\x14\x84\x10\x04\x59\xd8\x36\x48\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x85: bytearray(  # Mosfet status frame
            b"\x55\x14\x85\x08\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x87: bytearray(  # Temperatures frame
            b"\x55\x14\x87\x00\xe6\x00\xe6\x00\xe6\x00\xe6\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x88: bytearray(  # Cell voltages frame
            b"\x55\x14\x88\x0d\x57\x0d\x5b\x0d\x5b\x0d\x5e\x0d\x5d\x0d\x5e\x0d\x61\x0d\x60\xaa"
        ),
        0x89: bytearray(  # Cell voltages frame
            b"\x55\x14\x89\x0d\x5d\x0d\x61\x0d\x5d\x0d\x5d\x0d\x5e\x0d\x5c\x0d\x5e\x0d\x65\xaa"
        ),
        0x8A: bytearray(  # Cell voltages frame
            b"\x55\x14\x8a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x90: bytearray(
            b"\x55\x14\x90\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x91: bytearray(
            b"\x55\x14\x91\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x94: bytearray(
            b"\x55\x14\x91\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
        0x95: bytearray(
            b"\x55\x14\x91\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaa"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("ff02"):
            frame: Final[bytes] = bytes(data)
            if frame[0] == 0x55 and frame[-1] == 0xAA and frame[1] == 0x04:
                return self.RESP[frame[2]]

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert self._notify_callback, (
            "write to characteristics but notification not enabled"
        )

        self._notify_callback(
            "MockTianPwrBleakClient", self._response(char_specifier, data)
        )


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test TianPwr BMS data update."""

    patch_bleak_client(MockTianPwrBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockTianPwrBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "hw_version": "TP-LT55",
        "sw_version": "0.1.10",
    }


@pytest.fixture(
    name="wrong_response",
    params=[
        (bytearray(b"\x51\x14\x83" + bytes(16) + b"\xaa"), "wrong_SOF"),
        (bytearray(b"\x55\x14\x83" + bytes(16) + b"\xa1"), "wrong_EOF"),
        (bytearray(b"\x55\x14\x83" + bytes(17) + b"\xaa"), "wrong_length_max"),
        (bytearray(b"\x55\x14\x83" + bytes(15) + b"\xaa"), "wrong_length_min"),
        (bytearray(), "empty_frame"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout, wrong_response: bytearray
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockTianPwrBleakClient,
        "RESP",
        MockTianPwrBleakClient.RESP | {0x83: wrong_response},
    )

    patch_bleak_client(MockTianPwrBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


async def test_missing_message(
    monkeypatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test data up date with BMS returning no message type 83 but 90."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockTianPwrBleakClient,
        "RESP",
        MockTianPwrBleakClient.RESP
        | {
            0x83: bytearray(
                b"\x55\x14\x90\x00\x3c\x15\x62\x01\x18\x00\xe6\x00\xfa\x00\x00\x30\x30\x00\x64\xaa"
            )
        },
    )
    patch_bleak_client(MockTianPwrBleakClient)

    bms = BMS(generate_ble_device())

    with pytest.raises(ValueError, match="BMS data incomplete."):
        await bms.async_update()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\x55\x14\x84\x10\x04\x59\xd8\x36\x48\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xaa"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\x55\x14\x84\x10\x04\x59\xd8\x36\x48\x00\x00\x80\x00\x00\x00\x00\x00\x00\x00\xaa"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request):
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch, patch_bleak_client, problem_response: tuple[bytearray, str]
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockTianPwrBleakClient,
        "RESP",
        MockTianPwrBleakClient.RESP | {0x84: problem_response[0]},
    )

    patch_bleak_client(MockTianPwrBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == ref_value() | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 63),
    }

    await bms.disconnect()



"""Test the Humsienk BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.humsienk_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

_RESULT_DEFS: BMSSample = {
    "battery_level": 42,
    "battery_health": 100,
    "voltage": 13.130,
    "current": -1.15,
    "cycle_charge": 63.46,
    "design_capacity": 150,
    "cycles": 3,
    "cell_count": 4,
    "cell_voltages": [3.281, 3.291, 3.284, 3.28],
    "battery_charging": False,
    "power": -15.099,
    "cycle_capacity": 833.23,
    "runtime": 198657,
    "delta_voltage": 0.011,
    "temperature": 30.5,
    "temp_values": [
        35.0,
        26.0,
        26.0,
        26.0,
        44.0,
        26.0,
    ],
    "chrg_mosfet": True,
    "dischrg_mosfet": True,
    "balancer": False,
    "problem_code": 0,
    "problem": False,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockHumsienkBleakClient(MockBleakClient):
    """Emulate a Humsienk BMS BleakClient."""

    _RESP: dict[bytes, bytearray] = {
        b"\xaa\x00\x00\x00\x00": bytearray(b"\xaa\x00\x00\x00\x00"),  # init
        b"\xaa\x10\x00\x10\x00": bytearray(
            b"\xaa\x10\x03\x42\x4d\x43\xe5\x00"
        ),  # device type: BMC
        b"\xaa\x11\x00\x11\x00": bytearray(
            b"\xaa\x11\x0a\x42\x4d\x43\x2d\x30\x34\x53\x30\x30\x31\x62\x02"
        ),  # model: BMC-04S001
        b"\xaa\x20\x00\x20\x00": bytearray(
            b"\xaa\x20\x0f\x00\x00\x00\x00\x80\x00\x80\x00\x00\x00\x00\x00\x00\x00\x00\x2f\x01"
        ),  # MOSFET status [7], [9], balancer [8]
        b"\xaa\x21\x00\x21\x00": bytearray(
            b"\xaa\x21\x1a\x4a\x33\x00\x00\x82\xfb\xff\xff\x2a\x64\xe4\xf7\x00\x00\xf0\x49\x02\x00"
            b"\x03\x00\x23\x1a\x1a\x1a\x2c\x1a\x91\x08"
        ),  # 13.130V, ...
        b"\xaa\x22\x00\x22\x00": bytearray(
            b"\xaa\x22\x30\xd1\x0c\xdb\x0c\xd4\x0c\xd0\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd2\x03"
        ),  # cell voltages
        b"\xaa\x23\x00\x23\x00": bytearray(
            b"\xaa\x23\x04\xdc\xfb\xff\xff\xfc\x03"
        ),  # current redundant with 0x21
        b"\xaa\x58\x00\x58\x00": bytearray(
            b"\xaa\x58\x30\x04\x00\x98\x3a\x42\x0e\x02\x0d\x04\x00\xc4\x09\xf0\x0a\x04\x00\xf8\x2a"
            b"\x10\x00\x30\x75\x14\x00\x66\x02\x50\x00\x03\x0d\xd1\x0c\xab\x0a\xdd\x0a\x35\x0d\x03"
            b"\x0d\xe3\x09\x15\x0a\xc5\x0d\x1e\x00\x05\x0b"
        ),
        b"\xaa\xf5\x00\xf5\x00": bytearray(
            b"\xaa\xf5\x0c\x56\x30\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb9\x01"
        ),  # version info
    }

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None
        self._notify_callback(
            "MockHumsienkBleakClient", self._RESP.get(bytes(data), bytearray())
        )


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Humsienk BMS data update."""

    patch_bleak_client(MockHumsienkBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockHumsienkBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {"hw_version": "V02", "model": "BMC-04S001"}


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"\xbb\x00\x00\x00\x00",
        b"\xaa\x00\x00\x00\x01",
        b"\xaa\x00\x00\x00\x00\x00",
        b"\xaa\x10\x00\x10\x00",
        b"\xaa",
        b"",
    ],
    ids=["wrong_SOF", "wrong_CRC", "wrong_len", "wrong_type", "only_SOF", "empty"],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockHumsienkBleakClient,
        "_RESP",
        MockHumsienkBleakClient._RESP | {b"\xaa\x00\x00\x00\x00": wrong_response},
    )

    patch_bleak_client(MockHumsienkBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.mark.parametrize(
    ("problem_response", "expected"),
    [
        (b"\x00\x00\x00\x00\x80\x00\x80\x00\x00\x00\x00\x00\x00\x80\x00\xaf\x01", 0x0),
        (b"\x00\x00\x00\x00\x80\x00\x80\x00\x00\x00\x00\x01\x00\x00\x00\x30\x01", 0x0),
        (
            b"\x00\x00\x00\x00\x80\x00\x80\x80\x00\x00\x00\x00\x00\x00\x00\xaf\x01",
            2**31,
        ),
        (b"\x00\x00\x00\x00\x81\x00\x80\x00\x00\x00\x00\x00\x00\x00\x00\x30\x01", 0x1),
    ],
    ids=["disc_highcon", "discon_low", "high", "low"],
)
async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: bytes,
    expected: int,
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockHumsienkBleakClient,
        "_RESP",
        MockHumsienkBleakClient._RESP
        | {b"\xaa\x20\x00\x20\x00": bytearray(b"\xaa\x20\x0f" + problem_response)},
    )

    patch_bleak_client(MockHumsienkBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == _RESULT_DEFS | {"problem": True, "problem_code": expected}

    await bms.disconnect()



"""Test the Renogy BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.renogy_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 512  # ATT max is 512 bytes


def ref_value() -> BMSSample:
    """Return reference value for mock Renogy BMS."""
    return {
        "battery_charging": False,
        "battery_level": 97.2,
        "cell_voltages": [3.5, 3.3, 3.3, 3.3],
        "cell_count": 4,
        "current": -0.86,
        "cycle_capacity": 1321.92,
        "cycle_charge": 97.2,
        "cycles": 15,
        "delta_voltage": 0.2,
        "design_capacity": 100,
        "power": -11.696,
        "problem": False,
        "problem_code": 0,
        "runtime": 406883,
        "temp_values": [17.0, 17.0],
        "temp_sensors": 2,
        "temperature": 17.0,
        "voltage": 13.6,
        "chrg_mosfet": False,
        "dischrg_mosfet": False,
        "heater": False,
    }


BASE_VALUE_CMD: Final[bytes] = b"\x30\x03\x13\xb2\x00\x07\xa4\x8a"


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockRenogyBleakClient(MockBleakClient):
    """Emulate a Renogy BMS BleakClient."""

    RESP: dict[bytes, bytearray] = {
        b"\x30\x03\x13\x88\x00\x22\x45\x5c": bytearray(
            b"\x30\x03\x44\x00\x04\x00\x23\x00\x21\x00\x21\x00\x21\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\xaa\x00"
            b"\xaa\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x25\x74"
        ),  # cell count (4/16), cell [0.1 V] (3.5, 3.3, 3.3, 3.3), temp count (2/16), temp [0.1°C] (17,17)
        b"\x30\x03\x13\xf0\x00\x1c\x44\x95": bytearray(  # Serial, name, sw-version
            b"\x30\x03\x38\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\xc8\x32\x30\x32\x31\x30\x35"
            b"\x32\x36\x00\x00\x00\x00\x00\x00\x00\x00\x20\x20\x20\x20\x20\x20\x20\x20\x52\x42\x54"
            b"\x31\x30\x30\x4c\x46\x50\x31\x32\x2d\x42\x54\x20\x20\x30\x31\x30\x30\x55\x2f"
        ),  # 08È20210526        RBT100LFP12-BT  0100
        BASE_VALUE_CMD: bytearray(
            b"\x30\x03\x0e\xff\xaa\x00\x88\x00\x01\x7b\xb0\x00\x01\x86\xa0\x00\x0f\xe0\x1e"
        ),  # -0.86A, 13.6V, 97.2Ah [mAh], 100Ah [mAh], (15 cycles (generated))
        b"\x30\x03\x13\xec\x00\x08\x85\x5c": bytearray(
            b"\x30\x03\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x96\xa1"
        ),  # alarm flags
        b"\x30\x03\x14\x02\x00\x08\xe4\x1d": bytearray(
            b"\x30\x03\x10\x52\x42\x54\x31\x30\x30\x4c\x46\x50\x31\x32\x2d\x42\x54\x20\x20\x58\xbb"
        ),  # RBT100LFP12-BT
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, cmd: bytes
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffd1"):
            return bytearray()

        return self.RESP.get(cmd, bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None

        resp: bytearray = self._response(char_specifier, bytes(data))
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockRenogyBleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Renogy BMS data update."""

    patch_bleak_client(MockRenogyBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockRenogyBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "serial_number": "20210526",
        "name": "RBT100LFP12-BT",
        "sw_version": "0100",
    }


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            b"\x30\x03\x0c\x00\x00\x00\x88\x00\x01\x7b\xb0\x00\x01\x86\xa0\x0d\xeb",
            "wrong_CRC",
        ),
        (
            b"\x31\x03\x0c\x00\x00\x00\x88\x00\x01\x7b\xb0\x00\x01\x86\xa0\x0c\xeb",
            "wrong_SOF",
        ),
        (
            b"\x30\x03\x0d\x00\x00\x00\x88\x00\x01\x7b\xb0\x00\x01\x86\xa0\x0c\xeb",
            "wrong_length",
        ),
        (bytes(2), "critical_length"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request) -> bytearray:
    """Return faulty response frame."""
    return bytearray(request.param[0])


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockRenogyBleakClient,
        "RESP",
        MockRenogyBleakClient.RESP | {BASE_VALUE_CMD: wrong_response},
    )

    patch_bleak_client(MockRenogyBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockRenogyBleakClient,
        "RESP",
        MockRenogyBleakClient.RESP
        | {
            b"\x30\x03\x13\xec\x00\x08\x85\x5c": bytearray(
                b"\x30\x03\x10\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd6"
                b"\xd1"
            )
        },
    )

    patch_bleak_client(MockRenogyBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == ref_value() | {
        "problem": True,
        "problem_code": 0xFFFFFFFFFFFFFFFFFFFFFFFFFFF1,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "heater": True,
    }

    await bms.disconnect()



#!/usr/bin/env python3
"""Generates a markdown table of natively reported data from the BMSs.

Lines are BMS types (from aiobmsble/bms/*_bms.py)
Columns are the fields in `BMSSample` directly filled from received data
A "✓" means that the data is natively available.
A "." means that the data is not natively available, but all required fields for its calculation are available.
Empty means that the data is not available at all.
"""
import csv
from importlib import import_module
from pathlib import Path
import re
from types import ModuleType
from typing import Any, Final

from aiobmsble import BMSSample, BMSValue
from aiobmsble.basebms import BaseBMS

ALWAYS_CALC: Final[frozenset[str]] = frozenset({"problem"})
ROOT: Final = Path(__file__).parents[1]
BMS_DIR: Final = ROOT / "aiobmsble" / "bms"
CSV_FILE: Final = ROOT / "docs" / "available_bms_data.csv"
INIT: Final = "aiobmsble"


def get_bmssample_fields(init_module: str) -> tuple[BMSValue, ...]:
    """Return the field names of the BMSSample dataclass."""

    module: ModuleType = import_module(init_module)
    return tuple(module.BMSSample.__annotations__.keys())


def file_has_field(path: Path, field: str) -> bool:
    """Check if the given BMS module file references the specified field."""
    # Prefer checking the imported module/class for _FIELDS so that
    # inherited fields from a base BMS class (defined in another module)
    # are correctly detected. Fall back to source text search if import fails.
    try:
        module: ModuleType = import_module(f"aiobmsble.bms.{path.stem}")
        cls: Final[Any | None] = getattr(module, "BMS", None)
        if cls is not None:
            fields: Final[Any | None] = getattr(cls, "_FIELDS", None)
            if fields is not None:
                for item in fields:
                    # NamedTuple BMSDp has attribute `key`
                    if getattr(item, "key", None) == field:
                        return True
                    # Some modules use tuple/list with the key as first element
                    if (
                        isinstance(item, (tuple, list))
                        and len(item)
                        and item[0] == field
                    ):
                        return True
    except ModuleNotFoundError:
        pass

    txt: Final[str] = path.read_text(encoding="utf8")
    patterns: Final[list[str]] = [
        rf'^\s*BMSDp\(\s*["\']{re.escape(field)}["\']',  # field in _FIELDS via BMSDp(...)
        rf'\bresult\[\s*["\']{re.escape(field)}["\']\s*\]',  # result["field"]
        rf'\bdata\[\s*["\']{re.escape(field)}["\']\s*\]',  # data["field"]
        rf'["\']{re.escape(field)}["\']\s*:',  # dict literal 'field': ...
        rf'\s\("{re.escape(field)}", \d+, ',  # tuples ("field", pos, ...
        # rf"\b{re.escape(field)}\b",  # fallback bareword
    ]
    return any(re.search(p, txt, re.MULTILINE) for p in patterns)


def get_bms_info(path: Path) -> tuple[str, str]:
    """Extract default_manufacturer and default_model from BMS INFO class variable."""

    module: ModuleType = import_module(f"aiobmsble.bms.{path.stem}")
    return module.BMS.INFO.get("default_manufacturer"), module.BMS.INFO.get(
        "default_model"
    )


def main() -> None:
    """Generate and print a markdown table of BMS data availability."""

    fields: Final[tuple[BMSValue, ...]] = tuple(
        field for field in get_bmssample_fields(INIT)
    )
    bms_files: Final[list[Path]] = sorted(
        p for p in BMS_DIR.glob("*_bms.py") if p.is_file()
    )
    header: Final[tuple[str, ...]] = (
        "BMS",
        *(field.replace("_", " ") for field in fields),
    )
    rows: list[list[str]] = []

    calculations: Final = BaseBMS._calculation_registry(BMSSample())  # noqa: SLF001

    for f in bms_files:
        manufacturer, model = get_bms_info(f)
        bms_name: str = (
            f"{manufacturer} ({model})"
            if manufacturer and model
            else f.name.removesuffix("_bms.py")
        )
        marks: tuple[str, ...] = tuple(
            ("✓" if file_has_field(f, field) else "." if field in ALWAYS_CALC else "")
            for field in fields
        )

        available_fields: list[BMSValue] = [
            field for field in fields if file_has_field(f, field)
        ]
        for field, (required_fields, _calc_func) in calculations.items():
            if required_fields.issubset(available_fields):
                # If all required fields for a calculated field are available,
                # we consider the calculated field as available too
                available_fields.append(field)

        for field, mark in zip(fields, marks, strict=True):
            if mark:
                continue
            if field not in available_fields:
                continue
            marks = tuple(
                ("." if fld == field else mk)
                for fld, mk in zip(fields, marks, strict=True)
            )

        rows.append([bms_name, *marks])

    with Path.open(CSV_FILE, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(header)
        writer.writerows(rows)


if __name__ == "__main__":
    main()



"""Test the ANT implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.ant_leg_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_ADDRESS = "aa:bb:cc:a2:34:56"
BT_FRAME_SIZE: Final[int] = 19  # ANT BMS frame size

_RESULT_DEFS: Final[BMSSample] = {
    "voltage": 54.2,
    "current": -0.0,
    "battery_level": 100,
    "cycle_charge": 139.992578,
    "total_charge": 188250,
    "design_capacity": 140,
    "cycle_capacity": 7587.598,
    "cycles": 1344,
    "power": -0.0,
    "battery_charging": False,
    "cell_count": 16,
    "cell_voltages": [
        3.338,
        3.339,
        3.339,
        3.339,
        3.413,
        3.391,
        3.436,
        3.4,
        3.464,
        3.446,
        3.398,
        3.506,
        3.339,
        3.339,
        3.339,
        3.339,
    ],
    "delta_voltage": 0.168,
    "temp_values": [26, 29, -5, 21],
    "temperature": 17.75,
    "problem": True,
    "problem_code": 512,
    "balancer": False,
    "chrg_mosfet": False,
    "dischrg_mosfet": True,
    "runtime": 169081843,
}

_RESULT_DEFS_CAP: Final[BMSSample] = {
    "voltage": 48.8,
    "current": -8.0,
    "battery_level": 41,
    "design_capacity": 170,
    "cycle_charge": 68.769939,
    "total_charge": 11109,
    "runtime": 16386097,
    "cell_count": 14,
    "cell_voltages": [
        3.498,
        3.484,
        3.492,
        3.47,
        3.484,
        3.472,
        3.508,
        3.479,
        3.509,
        3.509,
        3.496,
        3.473,
        3.486,
        3.468,
    ],
    "temp_values": [22.0, 21.0, 21.0, 21.0],
    "delta_voltage": 0.041,
    "cycle_capacity": 3355.973,
    "cycles": 65,
    "power": -390.4,
    "battery_charging": False,
    "temperature": 21.25,
    "problem": False,
    "problem_code": 0,
    "balancer": False,
    "chrg_mosfet": True,
    "dischrg_mosfet": True,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockANTLEGACYBleakClient(MockBleakClient):
    """Emulate a ANT (legacy) BMS BleakClient."""

    CMDS: Final[dict[int, bytearray]] = {
        BMS.ADR.STATUS: bytearray(b"\xdb\xdb\x00\x00\x00\x00"),
    }
    RESP: Final[dict[int, bytearray]] = {
        BMS.ADR.STATUS: bytearray(
            b"\xaa\x55\xaa\xff\x02\x1e\x0d\x0a\x0d\x0b\x0d\x0b\x0d\x0b\x0d\x55\x0d\x3f\x0d\x6c\x0d"
            b"\x48\x0d\x88\x0d\x76\x0d\x46\x0d\xb2\x0d\x0b\x0d\x0b\x0d\x0b\x0d\x0b\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x00\x00\x00\x08\x58\x1e\x02\x0b"
            b"\x38\x79\x53\x0a\x13\xfb\xf3\x00\x1a\x00\x1d\xff\xfb\x00\x15\x00\x00\x00\x00\x02\x01"
            b"\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x0d\xb2\x01\x0d\x0a\x0d\x39\x10\xff\xef"
            b"\x00\x80\x00\x00\x00\x00\x00\x00\x0b\x50\x34\x09\x0f\x0d"
        ),
    }

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: bytearray = self.RESP.get(bytes(data)[2]) or bytearray()
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockANTBleakClient", notify_data)


async def test_update(
    patch_bms_timeout, patch_bleak_client, keep_alive_fixture
) -> None:
    """Test ANT BMS data update."""

    patch_bms_timeout()
    patch_bleak_client(MockANTLEGACYBleakClient)

    bms = BMS(
        generate_ble_device(BT_ADDRESS, "MockBLEdevice"),
        keep_alive_fixture,
    )

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_update_with_design_cap(
    monkeypatch, patch_bms_timeout, patch_bleak_client
) -> None:
    """Test ANT BMS data update with 0% battery SOC."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockANTLEGACYBleakClient,
        "RESP",
        MockANTLEGACYBleakClient.RESP
        | {
            BMS.ADR.STATUS: bytearray(
                b"\xaa\x55\xaa\xff\x01\xe8\x0d\xaa\x0d\x9c\x0d\xa4\x0d\x8e\x0d\x9c\x0d\x90\x0d\xb4"
                b"\x0d\x97\x0d\xb5\x0d\xb5\x0d\xa8\x0d\x91\x0d\x9e\x0d\x8c\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x50\x29\x0a\x21\xfe\x80\x04"
                b"\x19\x58\x93\x00\xa9\x84\x0f\x00\xfa\x08\x31\x00\x16\x00\x15\x00\x15\x00\x15\x00"
                b"\x15\x00\x15\x01\x01\x00\x03\xe8\x00\x17\x00\x00\x00\x01\x86\x09\x0d\xb5\x0e\x0d"
                b"\x8c\x0d\x9f\x0e\x00\x00\x00\x70\x00\x6b\x02\xac\x00\x00\x00\x00\x40\x01\x15\xf4"
            )
        },
    )

    patch_bleak_client(MockANTLEGACYBleakClient)

    bms = BMS(generate_ble_device(BT_ADDRESS, "MockBLEdevice"))

    assert await bms.async_update() == _RESULT_DEFS_CAP

    await bms.disconnect()


async def test_update_empty_battery(
    monkeypatch, patch_bms_timeout, patch_bleak_client
) -> None:
    """Test ANT BMS data update with 0% battery SOC."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockANTLEGACYBleakClient,
        "RESP",
        MockANTLEGACYBleakClient.RESP
        | {
            BMS.ADR.STATUS: bytearray(
                b"\xaa\x55\xaa\xff\x02\x1e\x0d\x0a\x0d\x0b\x0d\x0b\x0d\x0b\x0d\x55\x0d\x3f\x0d\x6c"
                b"\x0d\x48\x0d\x88\x0d\x76\x0d\x46\x0d\xb2\x0d\x0b\x0d\x0b\x0d\x0b\x0d\x0b\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08"
                b"\x58\x1e\x02\x0b\x38\x79\x53\x0a\x13\xfb\xf3\x00\x1a\x00\x1d\xff\xfb\x00\x15\x00"
                b"\x00\x00\x00\x02\x01\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x0d\xb2\x01\x0d"
                b"\x0a\x0d\x39\x10\xff\xef\x00\x80\x00\x00\x00\x00\x00\x00\x0b\x50\x34\x09\x0e\xa9"
            )
        },
    )

    patch_bleak_client(MockANTLEGACYBleakClient)

    bms = BMS(generate_ble_device(BT_ADDRESS, "MockBLEdevice"))

    _expected: BMSSample = _RESULT_DEFS.copy()
    _expected["battery_level"] = 0
    _expected.pop("design_capacity")
    _expected.pop("cycles")
    assert await bms.async_update() == _expected

    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (b"\x6e" + MockANTLEGACYBleakClient.RESP[BMS.ADR.STATUS][1:], "wrong_SOF"),
        (
            b"\xaa\x55\xaa\xfe" + MockANTLEGACYBleakClient.RESP[BMS.ADR.STATUS][4:],
            "unknown_type",
        ),
        (b"\xaa\x55\xaa", "too_short"),
        (
            MockANTLEGACYBleakClient.RESP[BMS.ADR.STATUS] + b"\xff",
            "too_long",
        ),
        (
            MockANTLEGACYBleakClient.RESP[BMS.ADR.STATUS][:-2] + b"\xff\xff",
            "wrong_CRC",
        ),
        (bytearray(1), "empty_response"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request):
    """Return faulty response frame."""
    return request.param[0]


async def test_invalid_response(
    monkeypatch, patch_bleak_client, patch_bms_timeout, wrong_response
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockANTLEGACYBleakClient,
        "RESP",
        MockANTLEGACYBleakClient.RESP | {BMS.ADR.STATUS: wrong_response},
    )

    patch_bleak_client(MockANTLEGACYBleakClient)

    bms = BMS(generate_ble_device(BT_ADDRESS, "MockBLEdevice"))

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()



"""Test the Daly BMS implementation."""

import asyncio
from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.ogt_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import DefGATTChar, MockBleakClient
from tests.test_basebms import BMSBasicTests

_RESULT_DEFS: Final[BMSSample] = {
    "voltage": 45.681,
    "battery_level": 14,
    "cycles": 99,
    "cycle_charge": 8.0,
    "temperature": 21.75,
    "temp_values": [21.75],
    "cycle_capacity": 365.448,
    "power": 56.188,
    "problem": False,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


# all names result in same encryption key for easier testing
@pytest.fixture(name="ogt_bms_name", params=["SmartBat-A12345", "SmartBat-B12294"])
def ogt_bms_fixture(request: pytest.FixtureRequest) -> str:
    """Return OGT smart BMS names."""
    assert isinstance(request.param, str)
    return request.param


class MockOGTBleakClient(MockBleakClient):
    """Emulate an OGT BMS BleakClient."""

    KEY = 0x10  # key used for decoding, constants are encrypted with this key!
    RESP_TYPE_A: dict[int, bytearray] = {
        0x02: bytearray(b" U  \x1d\x1a"),  # battery_level: 14
        0x04: bytearray(b'"  # Q\x1d\x1a'),  # cycle_charge: 8.0
        0x08: bytearray(b"'!R\"\x1d\x1a"),  # voltage: 45.681
        0x0C: bytearray(b"(% R\x1d\x1a"),  # temperature: 21.8
        0x10: bytearray(b"(%VV  \x1d\x1a"),  # current: -1.23
        0x18: bytearray(b"'(  \x1d\x1a"),  # runtime: 7200
        0x2C: bytearray(b"&#  \x1d\x1a"),  # cycles: 99
    }
    RESP_TYPE_B: dict[int, bytearray] = {
        0x08: bytearray(b"(% R\x1d\x1a"),  # temperature: 21.8
        0x09: bytearray(b"'!R\"\x1d\x1a"),  # voltage: 45.681
        0x0A: bytearray(b"'R   Q\x1d\x1a"),  # current: 1.23
        0x0D: bytearray(b" U  \x1d\x1a"),  # battery_level: 14
        0x0F: bytearray(b'"  # Q\x1d\x1a'),  # cycle_charge: 8.0
        0x12: bytearray(b"VVVV\x1d\x1a"),  # runtime: 65536 (inf)
        0x17: bytearray(b"&#  \x1d\x1a"),  # cycles: 99
        0x3F: bytearray(b"UQ S\x1d\x1a"),  # 3.306
        0x3E: bytearray(b"U) S\x1d\x1a"),  # 3.300
        0x3D: bytearray(b"U( S\x1d\x1a"),  # 3.304
        0x3C: bytearray(b"U' S\x1d\x1a"),  # 3.303
    }

    async def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff6"):
            assert self._ble_device.name is not None
            if self._ble_device.name[9] == "A":
                assert (
                    bytearray(data)[0:4] == b";BQQ"
                ), "BMS type A command header incorrect."
            else:
                assert (
                    bytearray(data)[0:4] == b";B!&"
                ), "BMS type B command header incorrect."

            reg: Final[int] = int(
                bytearray((bytearray(data)[x] ^ self.KEY) for x in range(4, 6)).decode(
                    encoding="ascii"
                ),
                16,
            )
            assert self._ble_device.name is not None

            if self._ble_device.name[9] == "A" and reg in self.RESP_TYPE_A:
                return bytearray(b";BT<") + bytearray(data)[4:6] + self.RESP_TYPE_A[reg]

            if self._ble_device.name[9] == "B" and reg in self.RESP_TYPE_B:
                return bytearray(b";BT<") + bytearray(data)[4:6] + self.RESP_TYPE_B[reg]

            return bytearray(b";BT<") + bytearray(b"Ubb\x7f\x10")  # Error

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        assert self._notify_callback is not None
        value: Final[bytearray] = await self._response(char_specifier, data)

        asyncio.get_running_loop().call_soon(
            self._notify_callback, DefGATTChar(1, "fff4"), value
        )


async def test_update(patch_bleak_client, ogt_bms_name, keep_alive_fixture) -> None:
    """Test OGT BMS data update."""

    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(
        generate_ble_device("cc:cc:cc:cc:cc:cc", ogt_bms_name), keep_alive_fixture
    )

    result: BMSSample = await bms.async_update()

    # verify all sensors are reported
    if str(ogt_bms_name)[9] == "A":
        assert result == _RESULT_DEFS | {
            "current": -1.23,
            "power": -56.188,
            "battery_charging": False,
            "runtime": 7200,
        }
    else:
        assert result == _RESULT_DEFS | {
            "current": 1.23,
            "delta_voltage": 0.003,
            "power": 56.188,
            "battery_charging": True,
            "cell_count": 4,
            "cell_voltages": [3.306, 3.305, 3.304, 3.303],
        }

    # query again to check already connected state
    result = await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_update_16s(monkeypatch, patch_bleak_client) -> None:
    """Test OGT BMS data update for 16 cell (max possible)."""

    monkeypatch.setattr(
        MockOGTBleakClient,
        "RESP_TYPE_B",
        MockOGTBleakClient.RESP_TYPE_B
        | {
            key: bytearray(f"U{(chr(0x27),'(',')','Q')[key % 4]} S\x1d\x1a", "ascii")
            for key in range(0x30, 0x3C)
        },
    )
    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", "SmartBat-B12294"), False)

    # verify all sensors are reported
    assert await bms.async_update() == _RESULT_DEFS | {
        "current": 1.23,
        "delta_voltage": 0.003,
        "power": 56.188,
        "battery_charging": True,
        "cell_count": 16,
        "cell_voltages": [
            3.306,
            3.305,
            3.304,
            3.303,
            3.306,
            3.305,
            3.304,
            3.303,
            3.306,
            3.305,
            3.304,
            3.303,
            3.306,
            3.305,
            3.304,
            3.303,
        ],
    }


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockOGTBleakClient)
    bms = BMS(generate_ble_device(name="SmartBat-B15051"))
    assert await bms.device_info() == {"serial_number": "15051"}
    assert bms.name == "SmartBat-B15051"


@pytest.fixture(
    name="wrong_response",
    params=[
        (bytearray(7), "critical_length"),
        (bytearray(b";AT< )'!R\"\x1d\x1a"), "wrong_SOP"),
        (bytearray(b";BT< )'!R\""), "wrong_EOP"),
        (bytearray(b";BT<#RUN S\x1d\x1a"), "invalid_character"),
        (bytearray(b";BT<Ubb\x7f\x10"), "BMS_error"),
        (bytearray(b"invalid\xf0value"), "invalid_value"),
        (bytearray(b";BT<UQ S\x1d\x1a"), "wrong_reg"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    async def patch_resp(
        _self, char_specifier: BleakGATTCharacteristic | int | str | UUID, _data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff6"):
            return wrong_response
        raise NotImplementedError("wrong GATT characteristic")

    monkeypatch.setattr(MockOGTBleakClient, "_response", patch_resp)

    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", "SmartBat-B12294"))

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


async def test_invalid_bms_type(patch_bleak_client) -> None:
    """Test BMS with invalid type 'C'."""

    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(
        generate_ble_device("cc:cc:cc:cc:cc:cc", "SmartBat-C12294"), keep_alive=True
    )

    result: BMSSample = await bms.async_update()
    assert not result
    assert bms._client.is_connected
    await bms.disconnect()



"""Test the Felicity implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.felicity_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 35

RESP_VALUE: Final[dict[str, bytearray]] = {
    "dat": bytearray(
        b'{"CommVer":1,"wifiSN":"F100011002424470238","iotType":3,"dateTime":"20210101010459",'
        b'"timeZMin":480}'
    ),
    "rt": bytearray(
        b'{"CommVer":1,"wifiSN":"F100011002424470238","modID":1,"date":"20210101010501",'
        b'"DevSN":"100011002424470238","Type":112,"SubType":7300,"Estate":960,"Bfault":0,'
        b'"Bwarn":0,"Bstate":960,"BBfault":0,"BBwarn":0,"BTemp":[[130,130],[256,256]],"Batt":'
        b'[[52800],[-1],[null]],"Batsoc":[[3300,1000,300000]],"Templist":[[130,130],[0,0],'
        b'[65535,65535],[65535,65535]],"BattList":[[52750,65535],[-1,-1]],"BatsocList":'
        b'[[3300,1000,300000]],"BatcelList":[[3296,3296,3297,3297,3297,3297,3297,3297,3297,'
        b"3297,3296,3297,3297,3297,3297,3297],[65535,65535,65535,65535,65535,65535,65535,"
        b'65535,65535,65535,65535,65535,65535,65535,65535,65535]],"EMSpara":[[1,2]],"BMaxMin":'
        b'[[3297,3296],[2,0]],"LVolCur":[[576,480],[1500,1500]],"BMSpara":[[1,2]],"BLVolCu":'
        b'[[576,480],[1500,1500]],"BtemList":[[130,130,130,130,32767,32767,32767,32767]]}'
    ),
    "bas": bytearray(
        b'{"CommVer":1,"version":"2.06","wifiSN":"F100011002424470238","COM":3,"iotType":3,'
        b'"modID":1,"DevSN":"100011002424470238","Type":112,"SubType":7300,"DSwVer":65535,'
        b'"M1SwVer":519,"M2SwVer":16,"DHwVer":0,"CtHwVer":0,"PwHwVer":65535}'
    ),
}


def ref_value() -> BMSSample:
    """Return reference value for mock Seplos BMS."""
    return {
        "voltage": 52.8,
        "current": -0.1,
        "battery_level": 33.0,
        "cycle_charge": 99.0,
        "temperature": 13.0,
        "cycle_capacity": 5227.2,
        "power": -5.28,
        "battery_charging": False,
        "cell_count": 16,
        "cell_voltages": [
            3.296,
            3.296,
            3.297,
            3.297,
            3.297,
            3.297,
            3.297,
            3.297,
            3.297,
            3.297,
            3.296,
            3.297,
            3.297,
            3.297,
            3.297,
            3.297,
        ],
        "temp_values": [
            13.0,
            13.0,
            13.0,
            13.0,
        ],
        "delta_voltage": 0.001,
        "runtime": 3564000,
        "problem": False,
        "problem_code": 0,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockFelicityBleakClient(MockBleakClient):
    """Emulate a Felicity BMS BleakClient."""

    HEAD_CMD: Final[int] = 0x7B
    TAIL_CMD: Final[int] = 0x7D
    CMDS: Final[dict[str, bytearray]] = {
        "dat": bytearray(b"wifilocalMonitor:get Date"),
        "bas": bytearray(b"wifilocalMonitor:get dev basice infor"),
        "rt": bytearray(b"wifilocalMonitor:get dev real infor"),
    }
    RESP: Final[dict[str, bytearray]] = RESP_VALUE

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:

        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("49535258-184d-4bd9-bc61-20c647249616"):
            for k, v in self.CMDS.items():
                if bytearray(data).startswith(v):
                    return self.RESP[k]

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data)

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: bytearray = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockFelicityBleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture) -> None:
    """Test Felicity BMS data update."""

    patch_bleak_client(MockFelicityBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockFelicityBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "fw_version": 519,
        "sw_version": "2.06",
        "model_id": 112,
        "serial_number": "100011002424470238",
    }


async def test_problem_response(monkeypatch: pytest.MonkeyPatch, patch_bleak_client) -> None:
    """Test Felicity BMS data update with problem response."""

    prb_resp: dict[str, bytearray] = RESP_VALUE.copy()
    prb_resp["rt"][146:166] = b'"Bfault":1,"Bwarn":10'  # patch problem codes

    patch_bleak_client(MockFelicityBleakClient)

    monkeypatch.setattr(MockFelicityBleakClient, "RESP", prb_resp)

    bms = BMS(generate_ble_device(), False)

    assert await bms.async_update() == ref_value() | {
        "problem": True,
        "problem_code": 11,
    }

    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (b'"CommVer":1,"wifiSN":"F100011002424470238"}', "invalid frame start"),
        (b'{"CommVer":1,"wifiSN":"F100011002424470238"', "invalid frame end"),
        (b'{"CommVer":2,"wifiSN":"F100011002424470238"}', "invalid protocol"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytes:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytes)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockFelicityBleakClient, "_response", lambda _s, _c_, d: wrong_response
    )
    patch_bleak_client(MockFelicityBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()



"""Test the Seplos v2 implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.seplos_v2_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 20
REF_VALUE: BMSSample = {
    "cell_count": 16,
    "temp_sensors": 6,
    "voltage": 53.0,
    "current": 6.85,
    "battery_level": 51.4,
    "battery_health": 100.0,
    "cycle_charge": 143.94,
    "cycles": 128,
    "temperature": 23.65,
    "cycle_capacity": 7628.82,
    "power": 363.05,
    "battery_charging": True,
    "cell_voltages": [
        3.312,
        3.313,
        3.313,
        3.313,
        3.313,
        3.312,
        3.313,
        3.315,
        3.311,
        3.312,
        3.313,
        3.313,
        3.313,
        3.312,
        3.313,
        3.313,
    ],
    "temp_values": [22.8, 22.2, 22.3, 23.2, 27.7, 23.7],
    "delta_voltage": 0.004,
    "pack_count": 1,
    "balancer": False,
    "chrg_mosfet": True,
    "dischrg_mosfet": True,
    "heater": False,
    "problem": False,
    "problem_code": 0,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockSeplosv2BleakClient(MockBleakClient):
    """Emulate a Seplos v2 BMS BleakClient."""

    HEAD_CMD = 0x7E
    TAIL_CMD = 0x0D
    PROTOCOL = 0x10
    RESP: dict[bytes, bytearray] = {
        b"\x00\x46\x61\x00\x01\x00\xf7\xc1": bytearray(  # get single machine data, address is incorrect but device also ignores it
            b"\x7e\x14\x02\x61\x00\x00\x6a\x00\x02\x10\x0c\xf0\x0c\xf1\x0c\xf1\x0c\xf1\x0c"
            b"\xf1\x0c\xf0\x0c\xf1\x0c\xf3\x0c\xef\x0c\xf0\x0c\xf1\x0c\xf1\x0c\xf1\x0c\xf0"
            b"\x0c\xf1\x0c\xf1\x06\x0b\x8f\x0b\x89\x0b\x8a\x0b\x93\x0b\xc0\x0b\x98\x02\xad"
            b"\x14\xb4\x38\x3a\x06\x6d\x60\x02\x02\x6d\x60\x00\x80\x03\xe8\x14\xbb\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x02\x03\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc1"
            b"\xd7\x0d"  # 53.00V, 6.85A, 143.94Ah, 51.4%, 128 cycles, 280.00Ah, 16 cells, 6 temp sensors
        ),
        b"\00\x46\x51\x00\x00\x3a\x7f": bytearray(  # manufacturer information
            b"\x7e\x14\x00\x51\x00\x00\x24\x43\x41\x4e\x3a\x50\x4e\x47\x5f\x44\x59\x45\x5f"
            b"\x4c\x75\x78\x70\x5f\x54\x42\x42\x31\x31\x30\x31\x2d\x53\x50\x37\x36\x20\x10"
            b"\x06\x01\x01\x46\x01\xa2\x6c\x0d"
        ),
        b"\x00\x46\x62\x00\x00\xa6\x8a": bytearray(  # GPD for main pack
            b"\x7e\x14\x00\x62\x00\x00\x30\x00\x00\x10\x0c\xf4\x0c\xee\x06\x0b\x93\x0b\x7f"
            b"\x0b\xb6\x0b\x8d\x00\xd7\x14\xb4\x11\x14\x07\x20\xd0\x02\x08\x20\xd0\x00\x71"
            b"\x03\xe8\x14\xb9\x07\x00\x02\x03\x08\x00\x00\x00\x00\x00\x00\x00\x00\x76\x31"
            b"\x0d"
        ),
        # return bytearray( # GPD for extension pack
        #     b"\x7e\x14\x00\x62\x00\x00\x28\x00\x00\x10\x00\x00\x00\x00\x06\x00\x00\x00\x00"
        #     b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x00\x00\x00\x00\x00\x00\x00\x00"
        #     b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0a\x4f\x0d"
        # )
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("ff02")
            and bytearray(data)[0] == self.HEAD_CMD
        ):
            if (
                bytes([self.HEAD_CMD, self.PROTOCOL]) == bytes(data)[0:2]
                and bytes(data)[-1] == self.TAIL_CMD
            ):
                return self.RESP.get(bytes(data)[2:-1], bytearray())

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: bytearray = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockSeplosv2BleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Seplos V2 BMS data update."""

    patch_bleak_client(MockSeplosv2BleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == REF_VALUE

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockSeplosv2BleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {"sw_version": "16.6", "model": "B1101-SP76"}


async def test_short_message(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test Seplos V2 BMS data update with short message (missing values)."""

    monkeypatch.setattr(
        MockSeplosv2BleakClient,
        "RESP",
        MockSeplosv2BleakClient.RESP
        | {
            b"\x00\x46\x61\x00\x01\x00\xf7\xc1": bytearray(
                b"\x7e\x14\x02\x61\x00\x00\x3c\x00\x02\x10\x0c\xf0\x0c\xf1\x0c\xf1\x0c\xf1\x0c"
                b"\xf1\x0c\xf0\x0c\xf1\x0c\xf3\x0c\xef\x0c\xf0\x0c\xf1\x0c\xf1\x0c\xf1\x0c\xf0"
                b"\x0c\xf1\x0c\xf1\x06\x0b\x8f\x0b\x89\x0b\x8a\x0b\x93\x0b\xc0\x0b\x98\x02\xad"
                b"\x14\xb4\x38\x3a\x06\x6d\x60\x02\x02\x6d\xe1\x06\x0d"
            )
        },
    )

    patch_bleak_client(MockSeplosv2BleakClient)

    bms = BMS(generate_ble_device(), False)

    result: BMSSample = {}
    with pytest.raises(ValueError, match="message too short to decode data"):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (b"\x7e\x14\x00\x51\x00\x00\x01\x00\x7a\xef\x00", "invalid frame end"),
        (b"\x7e\x10\x00\x51\x00\x00\x01\x00\xbb\x29\x0d", "invalid version"),
        (b"\x7e\x14\x00\x51\x80\x00\x01\x00\xa7\xd7\x0d", "error response"),
        (b"\x7e\x14\x00\x51\x00\x00\x01\x00\x7a\xee\x0d", "invalid CRC"),
        (b"\x7e\x14\x00\x51\x00\x00\x01\x00\x7a\xef\x0d\x00", "oversized frame"),
        (b"\x7e\x14\x00\x51\x00\x00\x02\x00\x7a\xef\x0d", "undersized frame"),
        (
            b"\x7e\x14\x00\x62\x00\x00\x30\x00\x00\x10\x0c\xf4\x0c\xee\x06\x0b\x93\x0b\x7f"
            b"\x0b\xb6\x0b\x8d\x00\xd7\x14\xb4\x11\x14\x07\x20\xd0\x02\x08\x20\xd0\x00\x71"
            b"\x03\xe8\x14\xb9\x07\x00\x02\x03\x08\x00\x00\x00\x00\x00\x00\x00\x00\x76\x31"
            b"\x0d",
            "wrong response code",
        ),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytes:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytes)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockSeplosv2BleakClient, "_response", lambda _s, _c, _d: wrong_response
    )

    patch_bleak_client(MockSeplosv2BleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.mark.parametrize(
    ("problem", "expected"),
    [
        (
            bytearray(
                b"\x7e\x14\x00\x62\x00\x00\x30\x00\x00\x10\x0c\xf4\x0c\xee\x06\x0b\x93\x0b\x7f"
                b"\x0b\xb6\x0b\x8d\x00\xd7\x14\xb4\x11\x14\x07\x20\xd0\x02\x08\x20\xd0\x00\x71"
                b"\x03\xe8\x14\xb9\x07\x00\x02\x03\x08\xff\xff\xff\xff\xff\xff\xff\xff\xd0\xd0"
                b"\x0d"
            ),
            0x7DFFFFFFFFFF,
        ),
        (
            bytearray(
                b"\x7e\x14\x03\x62\x00\x00\x30\x00\x00\x10\x0e\x03\x0d\x70\x06\x0b\xf5\x0b\xed"
                b"\x0c\x0f\x0b\xf9\xff\xa5\x16\x2f\x0a\xf0\x07\x0a\xf0\x03\xe7\x0a\xf0\x01\xbb"
                b"\x03\xe8\x16\x2b\x01\x00\x01\x03\x08\x00\x11\x00\x00\x00\x02\x00\x00\xd5\x8d"
                b"\x0d"
            ),
            0x20000,  # Charging over-temp protection
        ),
    ],
    ids=("all", "over-temp"),
)
# Alarm flags: Events 1-6, excluding 7-8
async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem: bytearray,
    expected: int,
) -> None:
    """Test data update with BMS returning invalid data (wrong CRC)."""

    monkeypatch.setattr(
        MockSeplosv2BleakClient,
        "RESP",
        MockSeplosv2BleakClient.RESP | {b"\x00\x46\x62\x00\x00\xa6\x8a": problem},
    )

    patch_bleak_client(MockSeplosv2BleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == REF_VALUE | {
        "problem": True,
        "problem_code": expected,
    }



"""Tests for Pro BMS plugin.

Note: Test values updated on 2025-08-06 to match corrected field offsets:
- All field offsets validated against 987 Pro BMS packets
- Power field: Confirmed at data_section offset 28 (packet bytes 32-35)
- Power values cross-validated with V×I calculation (0.019% average error)
"""

import asyncio
import contextlib
import logging

from bleak.backends.device import BLEDevice
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.pro_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

# Actual recorded packets from device logs
RECORDED_PACKETS: dict[str, bytearray] = {
    "init_response": bytearray.fromhex(
        "55aa080380aa01040000002c52"
    ),  # Initialization response packet fct. 0x04
    "data_discharging": bytearray.fromhex(
        "55aa2d0480aa01701c05000096090080e2000000ad19000033000000ca050000890c0000770b0000044e000082648e684000"
    ),  # fct. 0x70, 13.08V, -2.454A, 22.6°C, 51% SOC
    "data_charging": bytearray.fromhex(
        "55aa2d0480aa01703b05000066340000da00000057100000200000008601000029460000580a0000eb4600000bd98c68afff"
    ),  # fct. 0x70, 13.39V, 13.414A, 21.8°C, 32% SOC
    "data_with_protection": bytearray.fromhex(
        "55aa2d0480aa01701c05000096090081e2000000ad19000033000000ca050000890c0000770b0000044e000082648e684000"
    ),  # fct. 0x70, Same as discharging but with protection bit
    "data_zero_soc": bytearray.fromhex(
        "55aa2d0480aa01703b05000066340000da00000057100000000000008601000029460000580a0000eb4600000bd98c68afff"
    ),  # fct. 0x70, Same as charging but with 0% SOC
    "basic_machine_info": bytearray.fromhex(
        "55aa200380aa0140008000000002000000f7040000c800000004010000065eaf7a4ee0f700"
    ),  # fct. 0x40, SW version 0x0104 => 1.4
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockProBMSBleakClient(MockBleakClient):
    """Mock Pro BMS BleakClient for testing."""

    _init_packet: bytearray = RECORDED_PACKETS["init_response"]

    def __init__(
        self, address_or_ble_device, disconnected_callback=None, **kwargs
    ) -> None:
        """Initialize the mock client."""
        super().__init__(address_or_ble_device, disconnected_callback, **kwargs)
        self._test_packet: bytearray = RECORDED_PACKETS["data_discharging"]
        self._streaming_task: asyncio.Task[None] | None = None
        self._stop_streaming: bool = False

    def set_test_packet(self, packet: bytearray) -> None:
        """Set the packet to return."""
        self._test_packet = packet

    async def _stream_data(self) -> None:
        """Continuously stream data packets like a real device."""
        # Wait for initialization to complete and the 1-second sleep

        while not self._stop_streaming:
            if self._notify_callback and self._test_packet:
                self._notify_callback(None, self._test_packet)
            # Send data every 10ms (real 500ms) to speed up tests
            await asyncio.sleep(0.01)

    async def write_gatt_char(self, char_specifier, data, response=None):
        """Mock write to handle initialization and data requests."""
        await super().write_gatt_char(char_specifier, data, response)

        if data == BMS._HEAD + BMS._CMD_INIT:
            # Send initialization response
            if self._notify_callback:
                self._notify_callback(None, self._init_packet)

        elif data == BMS._HEAD + BMS._CMD_TRIGGER_DATA:
            # Start streaming data packets
            if not self._streaming_task:
                self._stop_streaming = False
                self._streaming_task = asyncio.create_task(self._stream_data())

    async def disconnect(self) -> None:
        """Stop streaming on disconnect."""
        self._stop_streaming = True
        if self._streaming_task:
            self._streaming_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._streaming_task
        await super().disconnect()


@pytest.mark.asyncio
async def test_async_update_discharging(patch_bleak_client) -> None:
    """Test async update with discharging data."""
    device: BLEDevice = generate_ble_device("AA:BB:CC:DD:EE:FF", "Pro BMS")
    mock_client: MockProBMSBleakClient = MockProBMSBleakClient(device)
    mock_client.set_test_packet(RECORDED_PACKETS["data_discharging"])
    patch_bleak_client(lambda *args, **kwargs: mock_client)

    bms = BMS(device)

    assert await bms.async_update() == {
        "voltage": 13.08,
        "current": -2.454,  # 0x96090080: discharge
        "temperature": 22.6,
        "temp_values": [22.6],
        "battery_level": 51,
        "battery_charging": False,
        "cycle_charge": 65.73,
        "cycle_capacity": 859.748,
        "power": -32.09,
        "runtime": 96425,
        "problem_code": 0,
        "problem": False,
    }

    await bms.disconnect()


@pytest.mark.asyncio
async def test_async_update_charging(patch_bleak_client) -> None:
    """Test async update with charging data."""
    device: BLEDevice = generate_ble_device("AA:BB:CC:DD:EE:FF", "Pro BMS")
    mock_client: MockProBMSBleakClient = MockProBMSBleakClient(device)
    mock_client.set_test_packet(RECORDED_PACKETS["data_charging"])
    patch_bleak_client(lambda *args, **kwargs: mock_client)

    bms = BMS(device)

    assert await bms.async_update() == {
        "voltage": 13.39,
        "current": 13.414,
        "temperature": 21.8,
        "temp_values": [21.8],
        "battery_level": 32,
        "battery_charging": True,
        "cycle_charge": 41.83,
        "cycle_capacity": 560.104,
        "power": 179.61,
        "problem_code": 0,
        "problem": False,
    }
    await bms.disconnect()


@pytest.mark.asyncio
async def test_async_update_with_protection(patch_bleak_client) -> None:
    """Test async update with protection status."""
    device: BLEDevice = generate_ble_device("AA:BB:CC:DD:EE:FF", "Pro BMS")
    mock_client: MockProBMSBleakClient = MockProBMSBleakClient(device)
    mock_client.set_test_packet(RECORDED_PACKETS["data_with_protection"])
    patch_bleak_client(lambda *args, **kwargs: mock_client)

    bms = BMS(device)

    assert await bms.async_update() == {
        "voltage": 13.08,
        "current": -2.454,  # 0x96090080: discharge
        "temperature": 22.6,
        "temp_values": [22.6],
        "battery_level": 51,
        "battery_charging": False,
        "cycle_charge": 65.73,
        "cycle_capacity": 859.748,
        "power": -32.09,
        "runtime": 96425,
        "problem_code": 1,  # 0x81 & 0x7F = 1
        "problem": True,
    }
    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockProBMSBleakClient, BMS)


@pytest.mark.asyncio
async def test_async_update_incomplete_data(
    patch_bleak_client, patch_bms_timeout
) -> None:
    """Test handling of incomplete data packet."""
    device: BLEDevice = generate_ble_device("AA:BB:CC:DD:EE:FF", "Pro BMS")
    mock_client: MockProBMSBleakClient = MockProBMSBleakClient(device)
    # Set incomplete data (too short)
    mock_client.set_test_packet(bytearray.fromhex("55aa0d0480aa0170"))
    patch_bms_timeout("pro_bms")
    patch_bleak_client(lambda *args, **kwargs: mock_client)

    bms = BMS(device)

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        await bms.async_update()  # Should return empty dict for incomplete data
    assert not result

    await bms.disconnect()


@pytest.mark.asyncio
async def test_async_update_already_streaming(patch_bleak_client) -> None:
    """Test async update when already streaming (second update)."""
    device: BLEDevice = generate_ble_device("AA:BB:CC:DD:EE:FF", "Pro BMS")
    mock_client: MockProBMSBleakClient = MockProBMSBleakClient(device)
    mock_client.set_test_packet(RECORDED_PACKETS["data_charging"])
    patch_bleak_client(lambda *args, **kwargs: mock_client)

    bms = BMS(device, keep_alive=True)

    # First update to initialize
    await bms.async_update()

    # Second update should reuse existing connection
    assert await bms.async_update() == {
        "voltage": 13.39,
        "current": 13.414,
        "temperature": 21.8,
        "temp_values": [21.8],
        "battery_level": 32,
        "battery_charging": True,
        "cycle_charge": 41.83,
        "cycle_capacity": 560.104,
        "power": 179.61,
        "problem_code": 0,
        "problem": False,
    }

    await bms.disconnect()


@pytest.mark.asyncio
async def test_async_update_no_data_after_init(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test when initialization completes but no valid data is received."""

    # Replace write_gatt_char to send init response after trigger command
    # This will set the event but not _streaming
    # FIXME! put into class
    async def mock_write(self, char, data, response=None) -> None:
        await MockBleakClient.write_gatt_char(self, char, data, response)
        assert self._notify_callback is not None
        if data == BMS._CMD_INIT:
            # Send initialization response
            self._notify_callback(None, RECORDED_PACKETS["init_response"])
        elif data == BMS._CMD_TRIGGER_DATA:
            # Wait for the 1-second sleep and event clear to complete
            async def send_wrong_packet() -> None:
                # Send another init response instead of data
                self._notify_callback(None, RECORDED_PACKETS["init_response"])

            # Store task reference to prevent garbage collection
            self._streaming_task = asyncio.create_task(send_wrong_packet())

    monkeypatch.setattr(MockProBMSBleakClient, "write_gatt_char", mock_write)

    patch_bms_timeout()
    patch_bleak_client(MockProBMSBleakClient)

    bms = BMS(generate_ble_device(name="Pro BMS"))

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    assert (
        not bms._client.is_connected
    ), "BMS should be disconnected if streaming is not working."


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"\x55\xaa\x07\x03\x80\xaa\x01\x04\x00\x00\x00\x2c\x52",
        RECORDED_PACKETS["data_zero_soc"],
        b"\x00\xaa\x08\x03\x80\xaa\x01\x04\x00\x00\x00\x2c\x52",
        b"\x55\xaa\x08\x03",
        b"",
    ],
    ids=[
        "wrong_length",
        "unexpected_RT_data",
        "invalid_header",
        "short_packet",
        "empty_packet",
    ],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(MockProBMSBleakClient, "_init_packet", wrong_response)
    patch_bleak_client(MockProBMSBleakClient)

    bms = BMS(generate_ble_device())

    logging.getLogger(__name__)
    result: BMSSample = {}
    with caplog.at_level(logging.DEBUG), pytest.raises(TimeoutError):
        result = await bms.async_update()
    assert not result
    assert "failed to initialize BMS connection" in caplog.text

    await bms.disconnect()



"""Test the CBT power VB series BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.cbtpwr_vb_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 32


def ref_value() -> BMSSample:
    """Return reference value for mock CBT power VB series BMS."""
    return {
        "voltage": 13.3,
        "current": 0.0,
        "battery_level": 96,
        "cycles": 0,
        "cycle_charge": 192.0,
        "cell_count": 4,
        "cell_voltages": [3.328, 3.326, 3.326, 3.326],
        "delta_voltage": 0.002,
        "design_capacity": 200,
        "temp_values": [6.2, 7.3],
        "temp_sensors": 2,
        "temperature": 6.75,
        "cycle_capacity": 2553.6,
        "power": 0.0,
        "battery_charging": False,
        "problem": False,
        "problem_code": 0,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockCBTpwrVBBleakClient(MockBleakClient):
    """Emulate a CBT power VB series BMS BleakClient."""

    _wr_buffer: bytearray = bytearray()  # collect individual write calls
    RESP: dict[bytes, bytearray] = {
        b"~11014642E00201FD35\r": bytearray(
            b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x36\x30\x34\x36\x30\x34\x30\x44\x30\x30\x30"
            b"\x43\x46\x45\x30\x43\x46\x45\x30\x43\x46\x45\x30\x32\x30\x30\x33\x45\x30\x30\x34"
            b"\x39\x30\x30\x30\x30\x30\x30\x38\x35\x30\x30\x36\x30\x30\x37\x30\x30\x30\x30\x30"
            b"\x30\x38\x36\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x45\x46\x36\x30\x0d"
        ),
        b"~11014681A006010001FC71\r": bytearray(
            b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x43\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
            b"\x46\x0d"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffe9"):
            return bytearray()

        return self.RESP.get(bytes(data), bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        if bytes(data).startswith(b"\x7e"):
            self._wr_buffer = bytearray(data)
        else:
            # concatenate write commands (BMS only accepts 20 bytes at once)
            self._wr_buffer.extend(bytes(data))
        if not bytes(data).endswith(b"\x0d"):
            return

        assert self._notify_callback is not None

        resp: Final[bytearray] = self._response(char_specifier, self._wr_buffer)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockCBTpwrVBBleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test CBT power VB series BMS data update."""

    patch_bleak_client(MockCBTpwrVBBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            bytearray(
                b"\x0e\x32\x32\x30\x31\x34\x36\x30\x30\x43\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x46\x0d"
            ),
            "wrong_SOF",
        ),
        (
            bytearray(
                b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x43\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x46\x0c"
            ),
            "wrong_EOF",
        ),
        (
            bytearray(
                b"\x7e\x29\x32\x30\x31\x34\x36\x30\x30\x43\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x46\x0d"
            ),
            "wrong_ENC",
        ),
        (
            bytearray(
                b"\x7e\x32\x31\x30\x31\x34\x36\x30\x30\x43\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x46\x0d"
            ),
            "wrong_VER",
        ),
        (
            bytearray(
                b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x43\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x30\x0d"
            ),
            "wrong_CRC",
        ),
        (
            bytearray(
                b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x30\x30\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x30\x0d"
            ),
            "wrong_LRC",
        ),
        (
            bytearray(
                b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x43\x29\x30\x34\x30\x37\x44\x30\x46\x43\x42"
                b"\x46\x0d"
            ),
            "wrong_head_ENC",
        ),
        (bytearray(13), "critical_length"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockCBTpwrVBBleakClient,
        "RESP",
        MockCBTpwrVBBleakClient.RESP | {b"~11014681A006010001FC71\r": wrong_response},
    )

    patch_bleak_client(MockCBTpwrVBBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x36\x30\x34\x36\x30\x34\x30\x44\x30\x30\x30"
                b"\x43\x46\x45\x30\x43\x46\x45\x30\x43\x46\x45\x30\x32\x30\x30\x33\x45\x30\x30\x34"
                b"\x39\x30\x30\x30\x30\x30\x30\x38\x35\x30\x30\x36\x30\x30\x37\x30\x30\x30\x30\x30"
                b"\x30\x38\x36\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x31\x45\x46\x35\x46\x0d"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\x7e\x32\x32\x30\x31\x34\x36\x30\x30\x36\x30\x34\x36\x30\x34\x30\x44\x30\x30\x30"
                b"\x43\x46\x45\x30\x43\x46\x45\x30\x43\x46\x45\x30\x32\x30\x30\x33\x45\x30\x30\x34"
                b"\x39\x30\x30\x30\x30\x30\x30\x38\x35\x30\x30\x36\x30\x30\x37\x30\x30\x30\x30\x30"
                b"\x30\x38\x36\x30\x30\x30\x30\x30\x30\x30\x30\x38\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x45\x46\x35\x38\x0d"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[dict[int, bytearray], str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[dict[int, bytearray], str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(  # patch response dictionary to only problem reports (no other data)
        MockCBTpwrVBBleakClient,
        "RESP",
        MockCBTpwrVBBleakClient.RESP | {b"~11014642E00201FD35\r": problem_response[0]},
    )

    patch_bleak_client(MockCBTpwrVBBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == ref_value() | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 47),
    }

    await bms.disconnect()



"""Test the Seplos BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.seplos_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 27  # ATT maximum is 512, minimal 27
CHAR_UUID = "fff1"
REF_VALUE: BMSSample = {
    "voltage": 52.34,
    "current": -6.7,
    "battery_level": 47.9,
    "battery_health": 99.9,
    "cycle_charge": 134.12,
    "cycles": 9,
    "temperature": 24.4,
    "cycle_capacity": 7019.841,
    "power": -350.678,
    "battery_charging": False,
    "runtime": 72064,
    "pack_count": 2,  # last packet does not report data!
    "cell_count": 16,
    "cell_voltages": [
        3.272,
        3.272,
        3.272,
        3.271,
        3.271,
        3.271,
        3.271,
        3.27,
        3.27,
        3.271,
        3.271,
        3.271,
        3.271,
        3.272,
        3.272,
        3.272,
        3.528,
        3.528,
        3.528,
        3.527,
        3.527,
        3.527,
        3.527,
        3.526,
        3.526,
        3.527,
        3.527,
        3.527,
        3.527,
        3.528,
        3.528,
        3.529,
    ],
    "delta_voltage": 0.003,
    "temp_values": [25.0, 23.8, 23.9, 24.9, 25.0, 23.8, 23.9, 24.9],
    "dischrg_mosfet": True,
    "chrg_mosfet": True,
    "balancer": False,
    "heater": False,
    "pack_battery_levels": [47.9, 48.0],
    "pack_battery_health": [99.9, 99.9],
    "pack_currents": [-7.2, -7.19],
    "pack_cycles": [9, 10],
    "pack_voltages": [52.34, 52.35],
    "problem": False,
    "problem_code": 0,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockSeplosBleakClient(MockBleakClient):
    """Emulate a Seplos BMS BleakClient."""

    PKT_FRAME = 0x5  # header(3) + crc(2)
    RESP: dict[str, bytearray] = {
        "EIA": bytearray(
            b"\x00\x04\x34\x14\x72\x00\x00\xff\xbd\xff\xff\x34\x64\x00\x00\x6d\x60\x00\x00\x00\xd5"
            b"\x00\x00\x6d\x60\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x07\x08\x00\x00\x07\x08\x00"
            b"\x00\x02\x40\x01\xd0\x00\x02\x00\x09\x01\xdf\x03\xe7\xb3\x36"  # A3F6
        ),
        "EIB": bytearray(
            b"\x00\x04\x2c\x0c\xc9\x0c\xc6\x00\x02\x00\x07\x14\x72\x14\x72\x00\x00\x00\x00\x00\xfa"
            b"\x00\xee\x00\xf4\x00\x00\x00\x01\x01\xdf\x01\xdf\x00\x09\x03\xe7\x01\x0a\x01\x0a\x01"
            b"\x0a\x00\x00\x00\x00\x57\x96"
        ),
        "EIC": bytearray(
            b"\x00\x01\x0a\x01\x00\x00\x00\x00\x00\x00\x03\x00\x00\x7e\x35"
        ),  # discharge, dis/charge-FET on, no alarm
        "EIX": bytearray(  # Note: unknown to the BMS implementation, just for testing
            b"\x00\x04\x20\x14\x72\xfd\x30\x34\x64\x6d\x60\x00\xd5\x01\xdf\x03\xe7\x00\x09\x0c\xc7"
            b"\x0b\x9f\x0c\xc8\x0c\xc6\x0b\xa5\x0b\x99\x00\x00\x00\xb4\x62\xb0"
        ),
        "PIA1": bytearray(
            b"\x01\x04\x22\x14\x72\xfd\x30\x34\x64\x6d\x60\x00\xd5\x01\xdf\x03\xe7\x00\x09\x0c\xc7"
            b"\x0b\x9f\x0c\xc8\x0c\xc6\x0b\xa5\x0b\x99\x00\x00\x00\xb4\x00\xb4\x6f\xf3"
        ),
        "PIA2": bytearray(
            b"\x02\x04\x22\x14\x73\xfd\x31\x34\x64\x6d\x60\x00\xd5\x01\xe0\x03\xe7\x00\x0a\x0c\xc7"
            b"\x0b\x9f\x0c\xc8\x0c\xc6\x0b\xa5\x0b\x99\x00\x00\x00\xb4\x00\xb4\xa6\xe2"
        ),
        "PIB1": bytearray(
            b"\x01\x04\x34\x0c\xc8\x0c\xc8\x0c\xc8\x0c\xc7\x0c\xc7\x0c\xc7\x0c\xc7\x0c\xc6\x0c\xc6"
            b"\x0c\xc7\x0c\xc7\x0c\xc7\x0c\xc7\x0c\xc8\x0c\xc8\x0c\xc8\x0b\xa5\x0b\x99\x0b\x9a\x0b"
            b"\xa4\x0a\xab\x0a\xab\x0a\xab\x0a\xab\x0b\xc4\x0b\xb5\x97\x1f"
        ),
        "PIB2": bytearray(
            b"\x02\x04\x34\x0d\xc8\x0d\xc8\x0d\xc8\x0d\xc7\x0d\xc7\x0d\xc7\x0d\xc7\x0d\xc6\x0d\xc6"
            b"\x0d\xc7\x0d\xc7\x0d\xc7\x0d\xc7\x0d\xc8\x0d\xc8\x0d\xc9\x0b\xa5\x0b\x99\x0b\x9a\x0b"
            b"\xa4\x0a\xab\x0a\xab\x0a\xab\x0a\xab\x0b\xc4\x0b\xb5\x53\xf1"
        ),
        "PIB3": bytearray(  # Note: incorrect answer for testing battery packet count mismatch
            b"\x02\x04\x34"
        )
        + bytearray(54),
    }

    def _crc16(self, data: bytearray) -> int:
        """Calculate CRC-16-CCITT XMODEM (ModBus)."""

        crc: int = 0xFFFF
        for i in data:
            crc ^= i & 0xFF
            for _ in range(8):
                crc = (crc >> 1) ^ 0xA001 if crc % 2 else (crc >> 1)
        return ((0xFF00 & crc) >> 8) | ((crc & 0xFF) << 8)

    def _response(self, data: Buffer) -> bytearray:

        req = bytearray(data)

        assert int.from_bytes(req[-2:]) == self._crc16(req[:-2])  # check CRC of request
        assert req[1] in [0x01, 0x04]  # check if read command

        device, start, length = [
            int(req[0]),
            int.from_bytes(req[2:4], byteorder="big"),
            int.from_bytes(req[4:6], byteorder="big") * (2 if req[1] == 0x4 else 0.125)
            + self.PKT_FRAME,
        ]

        if device == 0x00:  # EMS device
            if start == 0x2000:
                assert length == len(self.RESP["EIA"])
                return self.RESP["EIA"].copy()
            if start == 0x2100:
                assert length == len(self.RESP["EIB"])
                return self.RESP["EIB"].copy()
            if start == 0x2200:
                assert length == len(self.RESP["EIC"])
                return self.RESP["EIC"].copy()
        if device and device <= 0x10:  # BMS battery packs
            if start == 0x1000:
                assert length == len(self.RESP[f"PIA{device}"])
                return self.RESP[f"PIA{device}"].copy()
            if start == 0x1100:
                assert length == len(self.RESP[f"PIB{device}"])
                return self.RESP[f"PIB{device}"].copy()

        return bytearray()

    async def send_frag_response(
        self,
        data: Buffer,
        _response: bool | None = None,
    ) -> None:
        """Send fragmented response."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: bytearray = self._response(data)
        for notify_data in [
            resp[i : min(len(resp), i + BT_FRAME_SIZE)]
            for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockSeplosBleakClient", notify_data)

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"
        assert isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff2")

        await self.send_frag_response(data)


class MockWrongCRCBleakClient(MockSeplosBleakClient):
    """Emulate a Seplos BMS BleakClient returning wrong data."""

    async def send_frag_response(
        self,
        data: Buffer,
        _response: bool | None = None,
    ) -> None:
        """Send fragmented response."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: bytearray = self._response(data)
        resp[-2:] = bytearray(b"\00\00")  # make CRC invalid
        for notify_data in [
            resp[i : min(len(resp), i + BT_FRAME_SIZE)]
            for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockInvalidBleakClient", notify_data)


class MockErrRespBleakClient(MockSeplosBleakClient):
    """Emulate a Seplos BMS BleakClient returning error message."""

    def _response(self, data: Buffer) -> bytearray:
        # return error code 0x2 on read request (0x04)
        return bytearray(b"\x00\x84\x02\x93\x01")

    async def disconnect(self) -> None:
        """Mock disconnect to raise BleakError."""
        raise BleakError


class MockInvalidMessageBleakClient(MockSeplosBleakClient):
    """Emulate a Seplos BMS BleakClient returning unknown message type."""

    def _response(self, data: Buffer) -> bytearray:
        return self.RESP["EIX"]


class MockOversizedBleakClient(MockSeplosBleakClient):
    """Emulate a Seplos BMS BleakClient returning wrong data length."""

    async def send_frag_response(
        self,
        data: Buffer,
        _response: bool | None = None,
    ) -> None:
        """Send fragmented response and add trash to each message."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        # add garbage at the end for robustness
        resp = self._response(data) + bytearray(b"\xc0\xff\xee")

        for notify_data in [
            resp[i : min(len(resp), i + BT_FRAME_SIZE)]
            for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockOversizedBleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Seplos BMS data update."""

    patch_bleak_client(MockSeplosBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == REF_VALUE

    # query again to check already connected state
    assert await bms.async_update() == REF_VALUE
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockSeplosBleakClient, BMS)


async def test_wrong_crc(patch_bleak_client, patch_bms_timeout) -> None:
    """Test data update with BMS returning invalid data (wrong CRC)."""

    patch_bms_timeout()

    patch_bleak_client(MockWrongCRCBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


async def test_error_response(patch_bleak_client, patch_bms_timeout) -> None:
    """Test data update with BMS returning error message."""

    patch_bms_timeout()

    patch_bleak_client(MockErrRespBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


async def test_oversized_response(patch_bleak_client) -> None:
    """Test data update with BMS returning oversized data, result shall still be ok."""

    patch_bleak_client(MockOversizedBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == REF_VALUE

    await bms.disconnect()


async def test_invalid_message(patch_bleak_client, patch_bms_timeout) -> None:
    """Test data update with BMS returning error message."""

    patch_bms_timeout()
    patch_bleak_client(MockInvalidMessageBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


# Alaramflags used: TB02, TB03, TB05, TB06, TB15
#          skipped: TB09, TB04, TB16, TB07, TB08
async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test data update with BMS returning invalid data (wrong CRC)."""

    problem_resp: dict[str, bytearray] = MockSeplosBleakClient.RESP.copy()
    problem_resp["EIC"] = bytearray(
        b"\x00\x01\x0a\x01\xff\xff\xff\xff\xff\xff\x03\xff\xff\x4a\x75"
    )

    monkeypatch.setattr(MockSeplosBleakClient, "RESP", problem_resp)

    patch_bleak_client(MockSeplosBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == REF_VALUE | {
        "problem": True,
        "problem_code": 0xFFFF00FF00FF0000FF,
    }



"""Test the RoyPow BMS implementation."""

from asyncio import sleep
from collections.abc import Buffer
from copy import deepcopy
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.roypow_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 20
BT_MODULE_MSG: Final[bytes] = b"AT+STAT\r\n"  # AT cmd from BLE module


def ref_value() -> BMSSample:
    """Return reference value for mock Seplos BMS."""
    return {
        "temp_sensors": 4,
        "voltage": 13.48,
        "current": 0.35,
        "battery_level": 96,
        "cycle_charge": 98.056,
        "cycles": 2,
        "temperature": 19.25,
        "cycle_capacity": 1321.795,
        "power": 4.718,
        "battery_charging": True,
        "cell_count": 4,
        "cell_voltages": [3.375, 3.370, 3.369, 3.372],
        "temp_values": [19, 19, 19, 20],
        "delta_voltage": 0.006,
        "problem": False,
        "problem_code": 0,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockRoyPowBleakClient(MockBleakClient):
    """Emulate a RoyPow BMS BleakClient."""

    CMDS: Final[dict[int, bytearray]] = {
        0x02: bytearray(b"\xea\xd1\x01\x04\xff\x02\xf9\xf5"),
        0x03: bytearray(b"\xea\xd1\x01\x04\xff\x03\xf8\xf5"),
        0x04: bytearray(b"\xea\xd1\x01\x04\xff\x04\xff\xf5"),
    }
    RESP: Final[dict[int, bytearray]] = {
        0x02: bytearray(  # cell info
            BT_MODULE_MSG  # add BT message in front
            + b"\xea\xd1\x01\x0f\xff\x02\x04\x04\x04\x0d\x2f\x0d\x2a\x0d\x29\x0d\x2c\xf6\xf5"
        ),
        0x03: bytearray(  # +0.35A
            b"\xea\xd1\x01\x1a\xff\x03\x32\x00\x23\x00\x00\x00\x00\x04\x3b\x3b\x3b\x3c\x00\x10"
            b"\x00\x00\x00\x0c\x07\x00\x00\x00\xef\xf5"
            + BT_MODULE_MSG  # BT message at the end
        ),
        0x04: bytearray(  # 13.5V, 96%, 98.2Ah/105.0Ah
            b"\xea\xd1\x01\x39\xff\x04\x01\x60\x02\x00\x02\x03\x00\x01\x04\x9a\x28\x05\x00\x01\x06"
            b"\x9e\x24\x07\x00\x01\x08\x7f\x7d\x09\xff\xff\x0a\x04\xd4\x0b\x00\x16\x4a\x70\x0c\x25"
            b"\x03\x17\x00\x51\x45\x05\x44\x0d\x2e\x0d\x29\x0d\x14\x4e\x00\x00\x00\x6f\xf5"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:

        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("ffe1"):
            for k, v in self.CMDS.items():
                if bytearray(data).startswith(v):
                    return self.RESP[k]

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        self._notify_callback("MockRoyPowBleakClient", BT_MODULE_MSG)
        await sleep(0)

        resp: Final[bytearray] = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockRoyPowBleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test RoyPow BMS data update."""

    patch_bleak_client(MockRoyPowBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockRoyPowBleakClient, BMS)


async def test_update_dischrg(monkeypatch, patch_bleak_client) -> None:
    """Test RoyPow BMS data update."""

    patch_bleak_client(MockRoyPowBleakClient)

    negative_response: dict[int, bytearray] = deepcopy(MockRoyPowBleakClient.RESP)

    negative_response[0x3][6] |= 0x1  # make current negative
    negative_response[0x3][28] ^= 0x1  # patch CRC
    negative_response[0x4][30] = 0x0  # make runtime 255
    negative_response[0x4][-2] ^= 0xFF  # patch CRC

    monkeypatch.setattr(MockRoyPowBleakClient, "RESP", negative_response)

    bms = BMS(generate_ble_device(), False)

    assert await bms.async_update() == ref_value() | {
        "battery_charging": False,
        "current": -0.35,
        "power": -4.718,
        "runtime": 15300,
    }

    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            bytearray(
                b"\xee\xd1\x01\x0f\xff\x02\x04\x04\x04\x0d\x2f\x0d\x2a\x0d\x29\x0d\x2c\xf6\xf5"
            ),
            "wrong_SOF",
        ),
        (
            bytearray(
                b"\xea\xd1\x01\x0f\xff\x02\x04\x04\x04\x0d\x2f\x0d\x2a\x0d\x29\x0d\x2c\xf6\xe5"
            ),
            "wrong_EOF",
        ),
        (
            bytearray(
                b"\xea\xd1\x01\x0e\xff\x02\x04\x04\x04\x0d\x2f\x0d\x2a\x0d\x29\x0d\x2c\xf6\xf5"
            ),
            "wrong_length",
        ),
        (
            bytearray(
                b"\xea\xd1\x01\x0f\xff\x02\x04\x04\x04\x0d\x2f\x0d\x2a\x0d\x29\x0d\x2c\x00\xf5"
            ),
            "wrong_CRC",
        ),
        (bytearray(b"\x00\x00\x00\xf5"), "critical_length"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockRoyPowBleakClient,
        "RESP",
        MockRoyPowBleakClient.RESP | {0x2: wrong_response},
    )

    patch_bleak_client(MockRoyPowBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


async def test_missing_message(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test data up date with BMS returning no message type 4 but 8."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockRoyPowBleakClient,
        "RESP",
        MockRoyPowBleakClient.RESP
        | {
            0x4: bytearray(
                b"\xea\xd1\x01\x0f\xff\x08\x04\x04\x04\x0d\x2f\x0d\x2a\x0d\x29\x0d\x2c\xfc\xf5"
            )
        },
    )

    patch_bleak_client(MockRoyPowBleakClient)

    bms = BMS(generate_ble_device())

    with pytest.raises(ValueError, match="BMS data incomplete."):
        assert await bms.async_update() == {}
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\xea\xd1\x01\x1a\xff\x03\x32\x00\x23\x00\x00\x01\x00\x04\x3b\x3b\x3b\x3c\x00\x10"
                b"\x00\x00\x00\x0c\x07\x00\x00\x00\xee\xf5"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\xea\xd1\x01\x1a\xff\x03\x32\x00\x23\x80\x00\x00\x00\x04\x3b\x3b\x3b\x3c\x00\x10"
                b"\x00\x00\x00\x0c\x07\x00\x00\x00\x6f\xf5"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request):
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockRoyPowBleakClient,
        "RESP",
        MockRoyPowBleakClient.RESP | {0x3: problem_response[0]},
    )

    patch_bleak_client(MockRoyPowBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == ref_value() | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 23),
    }

    await bms.disconnect()



"""Module to support Ective BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
from string import hexdigits
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_sum


class BMS(BaseBMS):
    """Ective BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Topband", "default_model": "smart BMS"}
    _HEAD_RSP: Final[tuple[bytes, ...]] = (
        b"\x5e",
        b"\x83",
        b"\xb0",
    )  # header for responses
    _MAX_CELLS: Final[int] = 16
    _INFO_LEN: Final[int] = 113
    _CRC_LEN: Final[int] = 4
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 0, 4, False, lambda x: x / 1000),
        BMSDp("current", 4, 4, True, lambda x: x / 1000),
        BMSDp("battery_level", 14, 2, False),
        BMSDp("cycle_charge", 8, 4, False, lambda x: x / 1000),
        BMSDp("cycles", 12, 2, False),
        BMSDp("temp_values", 16, 2, False, lambda x: [round(x / 10 - 273.15, 3)]),
        BMSDp("problem_code", 18, 1, False),
    )

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: bytes = b""

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
                "manufacturer_id": m_id,
            }
            for m_id in (0, 0xFFFF)
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ffe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe4"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        raise NotImplementedError

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if (
            start := next(
                (i for i, b in enumerate(data) if bytes([b]) in BMS._HEAD_RSP), -1
            )
        ) != -1:  # check for beginning of frame
            data = data[start:]
            self._frame.clear()

        self._frame += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        if len(self._frame) < BMS._INFO_LEN:
            return

        del self._frame[BMS._INFO_LEN :]  # cut off exceeding data

        if not (
            self._frame.startswith(BMS._HEAD_RSP)
            and set(self._frame.decode(errors="replace")[1:]).issubset(hexdigits)
        ):
            self._log.debug("incorrect frame coding: %s", self._frame)
            self._frame.clear()
            return

        _dec: Final[bytes] = bytes.fromhex(
            self._frame.strip(b"".join(BMS._HEAD_RSP)).decode()
        )

        if (crc := crc_sum(_dec[:-2], 2)) != int.from_bytes(_dec[-2:]):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X", int.from_bytes(_dec[-2:]), crc
            )
            self._frame.clear()
            return

        self._msg = _dec
        self._msg_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)
        return self._decode_data(BMS._FIELDS, self._msg, byteorder="little") | {
            "cell_voltages": BMS._cell_voltages(
                self._msg, cells=BMS._MAX_CELLS, start=22, byteorder="little"
            )
        }



"""Test the ABC BMS implementation."""

import asyncio
from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.abc_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockABCBleakClient(MockBleakClient):
    """Emulate a ABC BMS BleakClient."""

    RESP: dict[int, bytearray] = {
        0xF0: bytearray(
            b"\xcc\xf0\xa2\x6b\x00\x00\x00\x00\xa0\x86\x01\x40\x9e\x01\x07\x00\x63\x00\x00\x21"
        ),
        0xF1: bytearray(
            b"\xcc\xf1\x53\x4f\x4b\x2d\x42\x4d\x53\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40"
        ),  # SOK-BMS
        0xF2: bytearray(
            b"\xcc\xf2\x01\x01\x01\x14\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x7f"
        ),
        0xF3: bytearray(
            b"\xcc\xf3\x17\x03\x12\x00\x64\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x36"
        ),
        0xF41: bytearray(
            b"\xcc\xf4\x01\x72\x0d\x00\x02\xa8\x0d\x00\x03\x2f\x0d\x00\x04\x88\x0d\x00\x00\x8b"
        ),  #           ^^ 1B idx, 3B voltage
        0xF45: bytearray(
            b"\xcc\xf4\x05\x81\x0d\x00\x06\x65\x0d\x00\x07\x1f\x0d\x00\x08\x5c\x0d\x00\x00\x33"
        ),
        0xF5: bytearray(
            b"\xcc\xf5\x42\x0e\x10\x0e\xc4\x09\xf6\x09\xa0\x86\x01\xa0\x86\x01\x00\x00\x00\x55"
        ),
        0xF6: bytearray(
            b"\xcc\xf6\x10\x72\x00\x80\x70\x00\x37\x00\x32\x00\x00\x00\x05\x00\x00\x00\x00\x23"
        ),
        0xF7: bytearray(
            b"\xcc\xf7\x20\x4e\x00\x40\x51\x00\x4b\x00\x46\x00\xec\xff\xf1\xff\x00\x00\x00\x8f"
        ),
        0xF8: bytearray(
            b"\xcc\xf8\x00\x64\x00\x80\x57\x00\x80\x70\x00\x10\x27\x00\x00\x00\x00\x00\x00\x3e"
        ),
        # 0xF9: bytearray(
        #     b"\xcc\xf9\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xed"
        # ),  # some BMS do not report this, so test without it (separate problem response test)
        0xFA: bytearray(
            b"\xcc\xfa\x48\x0d\x14\x00\x0f\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd6"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, cmd: int
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffe2"):
            return bytearray()

        return self.RESP.get(cmd, bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None

        for cmd in {  # determine which responses to command
            0xC0: [0xF1],
            0xC1: [0xF0, 0xF2],
            0xC2: [0xF0, 0xF3, 0xF41, 0xF45],
            0xC3: [0xF5, 0xF6, 0xF7, 0xF8, 0xFA],
            0xC4: [0xF9],
        }.get(bytearray(data)[1], []):
            self._notify_callback(
                "MockABCBleakClient", self._response(char_specifier, cmd)
            )
            await asyncio.sleep(0)


async def test_update(
    patch_bleak_client, patch_bms_timeout, keep_alive_fixture: bool
) -> None:
    """Test ABC BMS data update."""

    patch_bms_timeout()  # required for optional F9 response
    patch_bleak_client(MockABCBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == {
        "temp_sensors": 1,
        "voltage": 27.554,
        "current": 0.0,
        "battery_level": 99,
        "cycle_charge": 106.048,
        "cycles": 7,
        "cell_count": 8,
        "cell_voltages": [3.442, 3.496, 3.375, 3.464, 3.457, 3.429, 3.359, 3.42],
        "temp_values": [20],
        "delta_voltage": 0.137,
        "design_capacity": 100,
        "cycle_capacity": 2922.047,
        "power": 0.0,
        "battery_charging": False,
        "temperature": 20.0,
        "problem": False,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "heater": False,
    }

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(
        patch_bleak_client, MockABCBleakClient, BMS, {"model": "SOK-BMS"}
    )


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            b"\xcc\xf0\xa2\x6b\x00\x00\x00\x00\xa0\x86\x01\x40\x9e\x01\x07\x00\x63\x00\x00\x20",
            "wrong_CRC",
        ),
        (
            b"\xc0\xf0\xa2\x6b\x00\x00\x00\x00\xa0\x86\x01\x40\x9e\x01\x07\x00\x63\x00\x00\x21",
            "wrong_SOF",
        ),
        (
            b"\xcc\xfe\xa2\x6b\x00\x00\x00\x00\xa0\x86\x01\x40\x9e\x01\x07\x00\x63\x00\x00\x4f",
            "wrong_CMD",
        ),
        (
            b"\xcc\xf0\xa2\x6b\x00\x00\x00\x00\xa0\x86\x01\x40\x9e\x01\x07\x00\x63\x00\x21",
            "wrong_length",
        ),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    return bytearray(request.param[0])


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockABCBleakClient, "RESP", MockABCBleakClient.RESP | {0xF0: wrong_response}
    )

    patch_bleak_client(MockABCBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(ValueError, match="BMS data incomplete."):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            b"\xcc\xf9\x69\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1d",
            "first_bit",
        ),
        (
            b"\xcc\xf9\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x69\x00\x05",
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockABCBleakClient,
        "RESP",
        MockABCBleakClient.RESP | {0xF9: bytearray(problem_response[0])},
    )

    patch_bleak_client(MockABCBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result.get("problem", False)  # expect a problem report
    assert result.get("problem_code", 0) == (
        0x1 if problem_response[1] == "first_bit" else 0x8000
    )

    await bms.disconnect()
